/** ised - generating integer and floating point sequences
    Copyright (C) 2009-2011  Simon Copar

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/

#include "base.h"
#include "vm.h"
#include "methods.h"

#ifdef TIMERAND
#include <ctime>
#endif

#include <list>
#include <algorithm>
#include <cstring>
#include <cctype>

//if no history, no storing history
#ifndef HAVE_READLINE_HISTORY
#undef STOREHISTORY
#endif

#ifdef STOREHISTORY
#include<sys/stat.h>
#endif

using namespace std;

//functions

#define a  args[0]
#define a1 args[0]
#define a2 args[1]

void eval_control::evaluate_operator(uchar code, int argc, array* args, array& out){
  //static array of function pointers
  static value (value::*binops[8])(const value&) const={
    &value::operator+,
    &value::operator-,
    &value::operator*,
    &value::operator/,
    &value::operator%,
    &value::operator^,
    &value::e,
    &value::binomial 
  };
  static value (value::*logops[6])(const value&) const={
    &value::lt,
    &value::gt,
    &value::le,
    &value::ge,
    &value::eq,
    &value::neq
  };
  static void (value::*funmaps1[11])()={
    &value::op_fact,
    &value::op_sin,
    &value::op_cos,
    &value::op_tan,
    &value::op_asin,
    &value::op_acos,
    &value::op_atan,
    &value::op_sqrt,
    &value::op_cbrt,
    &value::op_exp,
    &value::op_log
  };

  static void (value::*funmaps2[7])()={
    &value::op_neg,
    &value::op_abs,
    &value::op_round,
    &value::op_frac,
    &value::op_int,
    &value::op_float,
    &value::op_rand
  };


  switch(code&OPMASK){
    case OPX_MEM:{
      switch(code){
	case OP_NONE: return;
	case OP_TAKE: {
	  int n=a.size();
	  for(int i=0;i<n;i++){
	    array* mem=0;
	    int index=a[i].assertI("${n}");
	    if(index<0){
	      if(-index<=histsize){
		mem=&memory[-histsize-1-index];
	      }else continue;
	    }else{
	      mem=&memory[index];
	    }
	    //copy because memory must survive
	    array memcopy(*mem);
	    out.append(memcopy);
	  }
	  return;
	}
	case OP_PUT:{
	  int n1=a1.size();
	  for(int i=0;i<n1;i++){
	    int index=a1[i].assertI("@{n}{...}");
	    if(index>=0)memory[index]=a2;
	  }
	  out.transfer(a2);
	  return;
	}
      }
      break;
    }
    case OPX_COM:{
      out.components(a1,a2,binops[(code&(~OPMASK))-1]);
      break;
    }
    case OPX_ACC:{
      switch(code){
	case OP_OPLUS:{
	  out.push_back(a.accum(value((itype)0),&value::operator+=));
	  return;
	}
	case OP_OTIMES:{
	  out.push_back(a.accum(value((itype)1),&value::operator*=));
	  return;
	}
	case OP_OMEAN:{
	  value v=a.accum(value((ieee)0),&value::operator+=);
	  v/=(itype)a.size();
	  out.push_back(v);
	  return;
	}
	case OP_OVAR:{
	  int n=a.size();
	  value v=(ieee)0;
	  value v2=(ieee)0;

	  for(array::iterator it=a.begin();it.ready();++it){
	    v+=(*it);
	    v2+=(*it)*(*it);
	  }
	  v/=(itype)n;
	  v2/=(itype)n;
	  v*=v;
	  out.push_back(v2-v);

	  return;
	}
	case OP_ONORM:{
	  int n1=a1.size();
	  out.allocate(n1);
	  a2.map(&value::op_abs);
	  for(array::iterator it1=a1.begin();it1.ready();++it1){
	    value v1=(*it1);
	    value v=(itype)0;
	    for(array::iterator it2=a2.begin();it2.ready();++it2){
	      v+=((*it2)^v1);
	    }
	    out.push_back(v^(ieee)(1/v1.d));
	  }
	  return;
	}
      }
      break;
    }
    case OPX_FUN1:{
      out.transfer(a);
      out.map(funmaps1[(code&(~OPMASK))-1]);
      break;
    }
    case OPX_FUN2:{
      out.transfer(a);
      out.map(funmaps2[(code&(~OPMASK))-1]);
      break;
    }
    case OPX_FUNX:{
      int n1=a1.size();
      int n2=a2.size();
      out.allocate(n1*n2);

      switch(code){
	case OP_BESJ:{
	  for(array::iterator it1=a1.begin();it1.ready();++it1){
	    int besn=(*it1).assertI("bj{n}{...}");
	    for(array::iterator it2=a2.begin();it2.ready();++it2){
	      out.push_back((*it2).op_bj(besn));
	    }
	  }
	  return;
	}
	case OP_BESY:{
	  for(array::iterator it1=a1.begin();it1.ready();++it1){
	    int besn=(*it1).assertI("by{n}{...}");
	    for(array::iterator it2=a2.begin();it2.ready();++it2){
	      out.push_back((*it2).op_by(besn));
	    }
	  }
	  return;
	}
      }
      break;
    }
    case OPX_TEN:{
      out.tensor(a1,a2,binops[(code&(~OPMASK))-1]);
      break;
    }
    case OPX_SET:{
      switch(code){
	case OP_AND:{
	  for(array::iterator it1=a1.begin();it1.ready();++it1){
	    value v1=(*it1);
	    if(a2.includes(v1))out.push_back(v1);
	  }
	  return;
	}
	case OP_WITHOUT:{
	  for(array::iterator it1=a1.begin();it1.ready();++it1){
	    value v1=(*it1);
	    if(!a2.includes(v1))out.push_back(v1);
	  }
	  return;
	}
	case OP_OR:{
	  out.append(a1);
	  out.append(a2);
	  return;
	}
	case OP_XOR:{
	  for(array::iterator it1=a1.begin();it1.ready();++it1){
	    value v1=(*it1);
	    if(!a2.includes(v1))out.push_back(v1);
	  }
	  for(array::iterator it2=a2.begin();it2.ready();++it2){
	    value v2=(*it2);
	    if(!a1.includes(v2))out.push_back(v2);
	  }
	  return;
	}
	case OP_REVERSE:{
	  out.transfer(a);
	  out.reverse();
	  return;
	}
	case OP_SORT:{
	  out.transfer(a);
	  out.sort(operator<);
	  return;
	}
	case OP_HASH:{
	  out.push_back((itype)a.size());
	  return;
	}
	case OP_MASK:{
	  int n=a.size();
	  array::iterator it=a.begin();
	  for(int i=0;i<n;i++){
	    if((*it)!=value((itype)0)){
	      out.push_back((itype)i);
	    }
	    ++it;
	  }
	  return;
	}
	case OP_EQUALS:{
	  int i=a1==a2?1:0;
	  out.push_back((itype)i);
	  return;
	}
	case OP_COUNT:{
	  for(array::iterator it1=a1.begin();it1.ready();++it1){
	    int count=0;
	    value x=(*it1);
	    for(array::iterator it2=a2.begin();it2.ready();++it2){
	      if(x==(*it2))count++;
	    }
	    out.push_back((itype)count);
	  }
	  return;
	}
	case OP_HIST:{
	  int n=a1.size();
	  out.allocate(n+1);
	  //initialize counts; bins are (-inf,a0),(a0,a1),...,(an-1,inf)
	  for(int i=0;i<=n;i++){
	    out.push_back((itype)0);
	  }
	  for(array::iterator it2=a2.begin();it2.ready();++it2){
	    int index=0;
	    value x=(*it2);
	    for(array::iterator it1=a1.begin();it1.ready();++it1){
	      if(x<(*it1))break;
	      index++;
	    }
	    out[index]+=(itype)1;
	  }
	  return;
	}

      }
      break;
    }
    case OPX_SEL:{
      switch(code){
	case OP_SHL:{
	  int n1=a1.size();
	  for(array::iterator it2=a2.begin();it2.ready();++it2){
	    int ihi=(*it2).assertI("{...}<<{n}");
	    if(ihi<0){//negative head is tail
	      ihi=-ihi;
	      for(int i=n1;i<ihi;i++)out.push_back((itype)0);//padding
	      array copy=a1;
	      copy.tail(n1-ihi);
	      out.append(copy);
	    }else{
	      array copy=a1;
	      copy.head(ihi);
	      out.append(copy);
	      for(int i=n1;i<ihi;i++)out.push_back((itype)0);//padding
	    }
	  }
	  return;
	}
	case OP_SHR:{
	  int n1=a1.size();
	  for(array::iterator it2=a2.begin();it2.ready();++it2){
	    int ihi=(*it2).assertI("{...}>>{n}");
	    if(ihi<0){//negative tail is head
	      ihi=-ihi;
	      array copy=a1;
	      copy.head(n1-ihi);
	      out.append(copy);
	      for(int i=n1;i<ihi;i++)out.push_back((itype)0);//padding
	    }else{
	      for(int i=n1;i<ihi;i++)out.push_back((itype)0);//padding
	      array copy=a1;
	      copy.tail(ihi);
	      out.append(copy);
	    }
	  }
	  return;
	}
	case OP_INDEX:{
	  int n1=a1.size();
	  int n2=a2.size();
	  if(!n1 || !n2)return;

	  out.allocate(n2);
	  for(array::iterator it2=a2.begin();it2.ready();++it2){
	    value v=(*it2);//%(itype)n1;
	    if(v.type==INTEGER){
	      out.push_back(a1[v.i]);
	    }else{
	      itype ii=v.intpart();
	      ieee f=v.d-ii;
	      if(f<0 || f>1){
		fprintf(stderr,"runtime: warning, array subscript (%g) out of integer range\n",v.d);
		continue;
	      }
	      out.push_back(a1[ii]*(1-f)+a1[ii+1]*f);
	    }
	  }
	  return;
	}
	case OP_UNIQUE:{
	  for(array::iterator it=a.begin();it.ready();++it){
	    value v1=(*it);
	    if(!out.includes(v1))out.push_back(v1);
	  }
	  return;
	}
	case OP_POSITIVE:{
	  a.filter(&value::is_positive,out);
	  return;
	}
	case OP_NEGATIVE:{
	  a.filter(&value::is_negative,out);
	  return;
	}
	case OP_ODD:{
	  a.filter(&value::is_odd,out);
	  return;
	}
	case OP_EVEN:{
	  a.filter(&value::is_even,out);
	  return;
	}
	case OP_INTEGER:{
	  a.filter(&value::is_int,out);
	  return;
	}
	case OP_DOUBLE:{
	  a.filter(&value::is_float,out);
	  return;
	}
	case OP_MIN:{
	  if(!a.size())return;
	  out.push_back(a.min(operator<));
	  return;
	}
	case OP_MAX:{
	  if(!a.size())return;
	  out.push_back(a.min(operator>));
	  return;
	}

      }
      break;
    }
    case OPX_CLOG:{
      out.components(a1,a2,logops[(code&(~OPMASK))-1]);
      break;
    }
    case OPX_TLOG:{
      out.tensor(a1,a2,logops[(code&(~OPMASK))-1]);
      break;
    }
    case OPX_CONST:{
      switch(code){
	case OP_PI:{
	  out.push_back(PI);
	  return;
	}
	case OP_AA:{
	  out.push_back(AA);
	  return;
	}
	case OP_DEG:{
	  out.push_back(DEG);
	  return;
	}
	case OP_EMC:{
	  out.push_back(EMC);
	  return;
	}
	case OP_PM:{
	  out.push_back((itype)1);
	  out.push_back((itype)(-1));
	  return;
	}
	case OP_MP:{
	  out.push_back((itype)(-1));
	  out.push_back((itype)1);
	  return;
	}
 
      }
      break;
    }
    case OPX_POLY:{
      switch(code){
	case OP_XPLUS:{
	  int n1=a1.size();
	  int n2=a2.size();
	  if(n1<n2){
	    out.transfer(a2);
	    array::iterator ito=out.begin();
	    for(array::iterator it=a1.begin();it.ready();++it){
	      (*ito)+=(*it);
	      ++ito;
	    }
	  }else{
	    out.transfer(a1);
	    array::iterator ito=out.begin();
	    for(array::iterator it=a2.begin();it.ready();++it){
	      (*ito)+=(*it);
	      ++ito;
	    }
	  }
	  return;
	}
	case OP_XMINUS:{
	  int n1=a1.size();
	  int n2=a2.size();
	  if(n1<n2){
	    out.transfer(a2);
	    out.map(&value::op_neg);
	    array::iterator ito=out.begin();
	    for(array::iterator it=a1.begin();it.ready();++it){
	      (*ito)+=(*it);
	      ++ito;
	    }
	  }else{
	    out.transfer(a1);
	    array::iterator ito=out.begin();
	    for(array::iterator it=a2.begin();it.ready();++it){
	      (*ito)-=(*it);
	      ++ito;
	    }
	  }
	  return;
	}
	case OP_XMUL:{
	  int n1=a1.size();
	  int n2=a2.size();
	  int n=n1+n2-1;
	  out.allocate(n);
	  for(int i=0;i<n;i++){
	    value v=(itype)0;
	    for(int j=0;j<=i;j++){
	      if((j<n1)&&(i-j<n2)){
		v+=a1[j]*a2[i-j];
	      }
	    }
	    out.push_back(v);
	  }
	  return;
	}
	case OP_XDIV:{
	  trimzeroes(a1);
	  trimzeroes(a2);
	  int n1=a1.size();
	  int n2=a2.size();
	  if((n1==0) || (n2==0)){
	    return;
	  }
	  int n=polydivmod(a1,a2,out);
	  out.tail(out.size()-n);
	  if(!out.size())out.push_back((itype)0);
	  return;
	}
	case OP_XMOD:{
	  trimzeroes(a1);
	  trimzeroes(a2);
	  int n1=a1.size();
	  int n2=a2.size();
	  if((n1==0) || (n2==0)){
	    return;
	  }
	  int n=polydivmod(a1,a2,out);
	  out.head(n);
	  if(!out.size())out.push_back((itype)0);
	  return;
	}

	case OP_PX:{
	  int n1=a1.size();
	  int n2=a2.size();
	  out.allocate(n2);
	  for(array::iterator it2=a2.begin();it2.ready();++it2){
	    value x=(*it2);
	    if(fabs(x.d)<1){
	      value v=a1[n1-1];
	      for(int j=n1-2;j>=0;j--){
		v*=x;
		v+=a1[j];
	      }
	      out.push_back(v);
	    }else{
	      value v=a1.front();
	      for(int j=1;j<n1;j++){
		v/=x;
		v+=a1[j];
	      }
	      v*=x^((itype)(n2-1));
	      out.push_back(v);
	    }
	  }
	  return;
	}
	case OP_PD:{
	  int n=a.size();
	  if(n<=1){
	    out.push_back((itype)0);
	    return;
	  }
	  out.allocate(n-1);
	  array::iterator it=a.begin();
	  ++it;
	  for(int i=1;i<n;i++){
	    out.push_back((*it)*(itype)i);
	    ++it;
	  }
	  return;
	}
	case OP_PZ:{
	  bairstow(a,out);
	  return;
	}

      }
      break;
    }
    case OPX_VEC:{
      switch(code){
	case OP_DOT:{
	  int n1=a1.size();
	  int n2=a2.size();
	  int n=n1<n2?n1:n2;
	  value v=(itype)0;
	  array::iterator it1=a1.begin();
	  array::iterator it2=a2.begin();
	  for(int i=0;i<n;i++){
	    v+=(*it1)*(*it2);
	    ++it1;
	    ++it2;
	  }
	  out.push_back(v);
	  return;
	}
	case OP_VLEN:{
	  int n=a.size();
	  value v=(itype)0;
	  for(array::iterator it=a.begin();it.ready();++it){
	    v+=(*it)*(*it);
	  }
	  v.op_sqrt();
	  out.push_back(v);
	  return;
	}
	case OP_VNRM:{
	  int n=a.size();
	  value v=(itype)0;
	  for(array::iterator it=a.begin();it.ready();++it){
	    v+=(*it)*(*it);
	  }
	  v.op_sqrt();
	  for(array::iterator it=a.begin();it.ready();++it){
	    (*it)/=v;
	  }
	  out.transfer(a);
	  return;
	}
	case OP_CROSS:{
	  int n1=a1.size();
	  int n2=a2.size();
	  int n=n1<n2?n1:n2;
	  out.allocate(n);
	  if(n==1){out.push_back((itype)0);return;}
	  for(int i=0;i<n;i++){
	    out.push_back(a1[i+1]*a2[i+2]-a1[i+2]*a2[i+1]);//automatic index wrapping
	  }
	  return;
	}

      }
      break;
    }
    case OPX_PAREN:{
      switch(code){

	case OP_LIST:{
	  for(int i=0;i<argc;i++){
	    out.append(args[i]);
	  }
	  return;
	}
	case OP_ITER1:{
	  array arg;
	  for(int i=0;i<argc;i++){
	    arg.append(args[i]);
	  }
	  int n=arg.size();
	  switch(n){
	    case 0:{
	      out.push_back((itype)42);
	      break;
	    }
	    case 1:{
	      int count=arg.front().assertI("[n]");
	      out.allocate(count);
	      for(int i=0;i<count;i++){
		out.push_back((itype)i);
	      }
	      break;
	    }
	    case 2:{
	      value v0=arg.front();
	      value v1=arg.back();
	      if(v1>v0){
		for(value v=v0;v.d<=v1.d;v+=(itype)1){
		  out.push_back(v);
		}
	      }else{
		for(value v=v0;v.d>=v1.d;v-=(itype)1){
		  out.push_back(v);
		}
	      }
	      break;
	    }
	    default:{
	      value v=arg.front();
	      value v1=arg.back();

	      //same value?

	      if(v.d==v1.d){
		out.push_back(v);
		break;
	      }
	      value max=(itype)0; //max. offset
	      value min=(itype)0;
	      value sum=(itype)0;
	      for(int i=1;i<n-1;i++){
		sum+=arg[i];
		if(sum>max)max=sum;
		if(sum<min)min=sum;
	      }
	      //is there minus?
	      if(v>v1){//if the sequence is descending or if it overshot final value
		if((sum.d<0)||(min<v1-v)){
		  int index=0;
		  for(int i=0;v.d>=v1.d;i++){
		    out.push_back(v);
		    v+=arg[1+index];
		    index++;
		    if(index>=(n-2))index-=(n-2);
		  }
		}
	      }else{//if the sequence is ascending or if it overshot final value
		if((sum.d>0)||(max>v1-v)){
		  int index=0;
		  for(int i=0;v.d<=v1.d;i++){
		    out.push_back(v);
		    v+=arg[1+index];
		    index++;
		    if(index>=(n-2))index-=(n-2);
		  }
		}
	      }
	      break;
	    }
	  }
	  return;
	}
	case OP_ITER2:{
	  array arg;
	  for(int i=0;i<argc;i++){
	    arg.append(args[i]);
	  }
	  int n=arg.size();
	  switch(n){
	    case 0:break;
	    case 1:{
	      out.push_back(arg.front());
	      break;
	    }
	    case 2:{
	      ieee limit=arg.back().d;
	      out.allocate((int)(limit)+1);
	      value v=arg.front();
	      for(int i=0;i<limit;i++){
		out.push_back(v);
	      }
	      break;
	    }
	    default:{
	      value v=arg.front();
	      ieee count=arg.back().d;//arg[n-1].d;
	      if(count<0)count=(-count)*(n-2)+1;//negative: whole sequence times count
	      out.allocate((int)(count)+1);
	      int index=0;
	      for(int i=0;i<count;i++){
		out.push_back(v);
		v+=arg[1+index];
		index++;
		if(index>=(n-2))index-=(n-2);
	      }
	      break;
	    }
	  }
	  return;
	}

      }
      break;
    }
      //end of outer switch
  }
}

#undef a
#undef a1
#undef a2

void eval_control::evaluate(treenode& node, array& out){
  switch(node.type){
    case NODE_OP:{
      int argc=node.childz.size();
      array* args=new array[argc];
      int count=0;
      for(list<treenode>::iterator i=node.childz.begin();i!=node.childz.end();i++){
#ifdef SIGHANDLE
        //interrupt
        if(shared::interrupt){return;}
#endif
        evaluate(*i,args[count]);//recursion!!!
	count++;
      }
      evaluate_operator(node.op.code,argc,args,out);
      //out.print(stderr,Fd,Ff);
      out.straighten();
      delete[] args;
      return;
    }
    case NODE_VAL:{
      out.push_back(node.val);
      return;
    }
  }
}


void eval_control::print_array(array& a, FILE* stream, const char* sep, char end){
    if(!a.size()){if(end)fputc(end,stream);return;}
    int count=a.size()-1;
    array::iterator it=a.begin();
    for(int i=0;i<count;i++){
      //a[i].print(stream,Fd,Ff);
      (*it).print(stream,Fd,Ff);
      ++it;
      if(sep)fputs(sep,stream);
#ifdef SIGHANDLE
      //interrupt - long outputs take forever, must be able to break loop
      if(shared::interrupt){fputc('\n',stream);return;}
#endif
    }
    a.back().print(stream,Fd,Ff);
    if(end)fputc(end,stream);
}

char* eval_control::strip(char* str){
  char* out=(char*)malloc(strlen(str)+1);
  char* pos=out;//current character
  char* end=out;//position of null
  char mode=0; //quote type
  for(;*str;str++){
    if(!mode){
      if(isspace(*str))continue;
      if(*str=='\''){mode='\'';continue;}
      if(*str=='"'){mode='"';continue;}
      if(*str=='\\'){mode=' ';str--;continue;}//we handle escapes later
      mode=' ';
      *pos=*str;pos++;end=pos;
      continue;
    }
    //escaped characters (not for '')
    if((*str=='\\')&&(mode!='\'')){
      switch(str[1]){
        case 0: break;
        case 't': *pos='\t';pos++;break;
        case 'n': *pos='\n';pos++;break;
        case 'r': *pos='\r';pos++;break;
        case 'f': *pos='\f';pos++;break;
        case 'b': *pos='\b';pos++;break;
        case 'v': *pos='\v';pos++;break;
        default: *pos=str[1];pos++;
      }
      //escaped characters count even at the end
      end=pos;
      if(str[1]){
        str++;
      }else{//null after escape
        fprintf(stderr,"input: escape character at the end of input\n");
      }
      continue;
    }
    if((mode!=' ')&&(*str==mode)){
      mode=' ';//reset mode
      break;
    }
    *pos=*str;pos++;
    //if no quotes, spaces at the end don't count
    if(!((mode==' ')&&isspace(*str)))end=pos;
  }
  if(!mode){//etirely empty line
    free(out);
    return 0;
  }
  if(mode!=' '){//there was no closing quote
   fprintf(stderr,"input: missing closing quote '%c'\n",mode);
  }
  *end=0;//put the null where the string ends
  return out;
}

FILE* eval_control::file_open(char* filename){
  if(!strcmp(filename,"-"))return stdin;
  FILE* f=fopen(filename,"r");
  if(f)return f;
  fprintf(stderr,"input: file '%s' could not be opened\n",filename);
  return 0;
}

void eval_control::evaluate_code(list<code_chunk>& code, bool printlast, array& out){
  int i=0;
  for(list<code_chunk>::iterator it=code.begin();it!=code.end();++it){
    array a;
    evaluate(it->code,a);
    if(i==code.size()-1 && it->printable){
      if(printlast){
	print_array(a,stdout,delim,'\n');
      }
      out.transfer(a);
      return;
    }
    //print if forced print is enabled or semicolon is missing
    if(it->printable || toggle_print)print_array(a,stdout,delim,'\n');
    i++;
  }
  return;
}


//no line comes out of precompile with newline in it!!!!
//returns the number of noncommented lines
int eval_control::precompile(char* line, char** comment,list<code_chunk>& code){
  //convert newline into a null
  char* newline=strchr(line,'\n');
  if(newline)*newline=0;

  //shabang lines should be ignored. This way you can call ised with
  //#!/usr/local/bin/ised --f
  char* shabang=strstr(line,"#!");
  //if line begins with #!, omit it.
  //if line is empty, omit it.
  if((shabang==line)||(*line==0)){if(comment)*comment=0;return 0;}
  //if there's no comment, comment is end of string.
  if(comment)*comment=shabang?shabang:strchr(line,0);
  if(shabang)*shabang=0;//premature end of line at comment

  char* end;
  while(end=strchr(line,';')){//semicolon separates chunks of code
#ifdef SIGHANDLE
    //interrupt
    if(shared::interrupt){if(comment)*comment=0;return 0;}
#endif
    *end=0;//fake end
    vector<token> tokens;
    tokenize(line,tokens);
    compiler c;
    code.push_back((code_chunk){c.compile(tokens),false});
    *end=';';//restore end
    line=end+1;
  }
  //something remains
  if(*line){
    //finalization
    vector<token> tokens;
    tokenize(line,tokens);
    compiler c;
    code.push_back((code_chunk){c.compile(tokens),true});
    if(shabang)*shabang='#';//restore original line
    return 1;
  }else{
    if(shabang)*shabang='#';//restore original line
    return 0;
  }
}

char* eval_control::formatdup(const char* fmt){
#if IVALUE==ILONG || VALUE==LONG
  char* out=(char*)malloc(strlen(fmt)+2);//it may only be one character longer
  char* out_chr=out;

  uint percent=0;
  for(const char* c=fmt;*c;c++){
    if(*c == '%'){
      //percent: 0 -> 1 (found percentage)
      //percent: 1 -> 0 (%% = single percentage as character)
      //percent: anything else -> anything else (if more than one format, don't do anything)
      percent^=0x1;
    }else{
      if(percent==1 && isalpha(*c)){//percent is closed by a specifier
#if IVALUE==ILONG
	if(strchr("dxXiu",*c)){*out_chr='l';out_chr++;}
#endif
#if IVALUE==ILONG && VALUE==LONG
	else
#endif
#if VALUE==LONG
	if(strchr("eEfgG",*c)){*out_chr='L';out_chr++;}
#endif
	//in case of l or L: already long specifier, no problem
	//if shouldn't be there, not our problem
	percent=2;//percent + alphanumeric means percent closed
      }
    }
    *out_chr=*c;
    out_chr++;
  }
  *out_chr=0;//null terminated string
  //fprintf(stderr,"format='%s'\n",out);
  return out;
#else
  return strdup(fmt);
#endif
}

char* eval_control::bufgets(FILE* file){
   char* out=0;
   //put some dummy non-null character
   buffer[bufsize-1]=' ';
   out=fgets(buffer,bufsize,file);
   //if it changed, the buffer was too short
   while(buffer[bufsize-1]!=' '){
     #if DEBUG>0
     fprintf(stderr,"input: buffer resize from %d to %d\n",bufsize,2*bufsize);
     #endif
     buffer=(char*)realloc(buffer,2*bufsize);
     buffer[2*bufsize-1]=' ';
     out=fgets(buffer+bufsize-1,bufsize+1,file);
     bufsize*=2;
   }
   return out;
}

#ifndef HAVE_LIBREADLINE
char* readline(const char* prompt){
  fputs(prompt,stdout);
  char* buf=(char*)malloc(BUF_SIZE);
  return fgets(buf,BUF_SIZE,stdin);
}
#endif


void eval_control::read_interactive(){
  credits_dump();
#ifdef STOREHISTORY
  char* home=getenv("HOME");
  char* hist_file=0;
  if(home){
    hist_file=(char*)malloc(strlen(home)+15);
    sprintf(hist_file,"%s/.ised_history",home);

    struct stat buf;
    if(!stat(hist_file,&buf)){//file exists
      if(int errno=read_history(hist_file)){
	fprintf(stderr,"readline: errno %d on history load\n",errno);
      }
    }
  }
#endif

#ifdef SIGHANDLE
  shared::wait_ptr=&awaiting;
#endif
  char prompt[7];
  strcpy(prompt,"ised> ");
  char* line;
  while(line=readline(prompt)){
#ifdef SIGHANDLE
    shared::state=STATE_WORK;
#endif
    //check for interactive interrupt
    if(!strcmp(line,"q")|| !strcmp(line,"quit"))break;
    input_line(line);
#ifdef SIGHANDLE
    shared::state=STATE_IDLE;
    shared::interrupt=0;
#endif
#ifdef HAVE_READLINE_HISTORY
    if(strlen(line))add_history(line);
#endif
    free(line);
    //change prompt:
    if(awaiting){
      sprintf(prompt,"%c= ",awaiting);
    }else{
      strcpy(prompt,"ised> ");
    }
  }
#ifdef SIGHANDLE
  shared::wait_ptr=0;
#endif

#ifdef STOREHISTORY
  if(home){
    if(int errno=write_history(hist_file)){
      fprintf(stderr,"readline: errno %d on history write\n",errno);
    }else{
      history_truncate_file(hist_file,256);
    }
  }
  if(hist_file)free(hist_file);
#endif

}



void eval_control::read_file(FILE* f){
  while(bufgets(f)){
    input_line(buffer);
  }
}


void eval_control::evaluate_line(char* line, char** comment, bool printlast, array& out){
  list<code_chunk> code;
  precompile(line,comment,code);
  out.clear();//not sure if necessary!
  evaluate_code(code,printlast,out);
}

void eval_control::load_file(FILE* f){
  char* comment;
  array a;
  while(bufgets(f)){
    evaluate_line(buffer,&comment,false,a);
    //save to memory unless it's commented or empty line
    if(comment)memory[++mempos].transfer(a);
  }
  memory[0].clear();
  memory[0].push_back((itype)mempos);
}

//returns 0 for normal code, 1 for nonprintable code and 2 for immediate return
int eval_control::retrieve_code(char* f, list<code_chunk>& code, bool file_or_line){
  char* comment;
  int outputlines=0;
  if(file_or_line){
    FILE* ffile=file_open(f);
    if(!ffile)return 2;
    while(bufgets(ffile)){
      outputlines=precompile(buffer,&comment,code);
    }

    fclose(ffile);
  }else{
    outputlines=precompile(f,&comment,code);
    if(!comment)return 2;//empty, failed or entirely commented
  }
  return outputlines==0?1:0;
}

void eval_control::apply_to_file(char* f, FILE* l, bool file_or_line){
  char* comment;

  list<code_chunk> code;
  int errcode=retrieve_code(f,code,file_or_line);
  if(errcode==2)return;
  int line=1;
  while(bufgets(l)){
    memory[0].clear();
    memory[0].push_back((itype)line);
    evaluate_line(buffer,&comment,false,memory[1]);
    if(comment)line++;
    if(comment){
      array a;
      evaluate_code(code,false,a);
      if(!errcode){//only print if there's something to print
        print_array(a,stdout,delim,0);
        if(*comment){
	  fputs(comment,stdout);
        }
	fputc('\n',stdout);
      }
    }else{//if empty, failed or entirely commented, preserve it (preserves blanks)
      if(!errcode){
	fputs(buffer,stdout);
	fputc('\n',stdout);
      }
    }
  }
}

void eval_control::interlace_code(list<code_chunk>& code){
  int codesize=code.size();
  array* temp_buffer=new array[codesize];
  int count=0;

  for(list<code_chunk>::iterator it=code.begin();it!=code.end();++it){
    evaluate(it->code,temp_buffer[count]);
    count++;
  }

  int index=0;
  while(1){
    array a;
    for(int j=0;j<codesize;j++){
      if(temp_buffer[j].size()>index){
        a.push_back(temp_buffer[j][index]);
      }
    }
    if(a.size()>0){
      print_array(a,stdout,delim,'\n');
      index++;
    }else break;
  }
  delete[] temp_buffer;
}

void eval_control::interlace_to_line(char* line){
  list<code_chunk> code;
  precompile(line,0,code);
  interlace_code(code);
}


void eval_control::interlace_to_file(char* f, FILE* l){
  list<code_chunk> code;
  precompile(f,0,code);

  char* comment;

  int line=1;
  while(bufgets(l)){
    memory[0].clear();
    memory[0].push_back((itype)line);
    evaluate_line(buffer,&comment,false,memory[1]);
    if(comment){
      line++;
      interlace_code(code);
    }else{
      fputs(buffer,stdout);
      fputc('\n',stdout);//this was somehow missing!
    }
  }
}

void eval_control::memory_dump(){
  typedef map<int,array>::iterator iterator;
  fprintf(stderr,"memory page:\n");
  for(iterator i=memory.begin();i!=memory.end();i++){
    fprintf(stderr,"@%+6d | ",(*i).first);
    print_array((*i).second,stderr," ",'\n');
  }
  fprintf(stderr,"end of memory page\n");
}

void eval_control::credits_dump(){
fprintf(stderr,"%s  Copyright (C) 2009-2011 Simon Copar\n\
This program comes with ABSOLUTELY NO WARRANTY.\n\
This is free software, and you are welcome to redistribute it\n\
under certain conditions.\n",
PACKAGE_STRING);
}

static const char* STR_DISABLED="DISABLED";
static const char* STR_ENABLED="ENABLED";

void eval_control::configure_dump(){
#if VALUE==FLOAT
const char* dstr="FLOAT";
#elif VALUE==DOUBLE
const char* dstr="DOUBLE";
#elif VALUE==LONG
const char* dstr="LONG DOUBLE";
#endif

#if IVALUE==IINT
const char* istr="INT";
#elif IVALUE==ILONG
const char* istr="LONG LONG";
#endif
fprintf(stderr,"Build parameters:\n\
  floating point precision=%s (%d bytes)\n\
  integer precision=%s (%d bytes)\n",
  dstr,sizeof(ieee),istr,sizeof(itype));
#ifndef HAVE_LIBREADLINE
 const char* lrlstr=STR_DISABLED;
#else
 const char* lrlstr=STR_ENABLED;
#endif
#ifndef HAVE_READLINE_HISTORY
 const char* rlhstr=STR_DISABLED;
#else
 const char* rlhstr=STR_ENABLED;
#endif
#ifndef STOREHISTORY
 const char* rlhstore=STR_DISABLED;
#else
 const char* rlhstore=STR_ENABLED;
#endif
#ifndef TRAP_OVR
 const char* ovrflw=STR_DISABLED;
#else
 const char* ovrflw=STR_ENABLED;
#endif
 fprintf(stderr,"  libreadline %s, history %s, history store %s, overflow handling %s\n", lrlstr,rlhstr,rlhstore,ovrflw);

}

void eval_control::input_line(char* line){
  char* tofree=0;
  if(!awaiting){
    if(!strncmp(line,"--m",3)){memory_dump();return;}
    if(!strncmp(line,"--p",3)){toggle_print=1;return;}
    if(!strncmp(line,"--n",3)){toggle_print=0;return;}
    if(!strncmp(line,"--v",3)){
      credits_dump();
      putc('\n',stderr);
      configure_dump();
      return;
    }
    const char* chars="faldtDFx";
    char compare[4]="---";

    while(*chars){
      compare[2]=*chars;
      if(!strncmp(line,compare,3)){
	tofree=line=strip(line+3);
        awaiting=*chars;
        break;
      }
      chars++;
    }
  }

  //waiting for something
  if(awaiting){
    if((!tofree)&&line){
      if(awaiting=='t'){
	tofree=line=strdup(line);
      }else{
	tofree=line=strip(line);
      }
    }
    if(!line)return;//really empty line
    switch(awaiting){
      case 'f':{
        awaiting=0;
	if(lfile){//if we want to apply to another file
	  FILE* l=file_open(lfile);
	  if(l){
	    if(xltoggle==XL_LINE){
	      apply_to_file(line,l,true);
	    }else{
	      extended_to_file(line,l,true);
	    }
	    fclose(l);
	  }
	  free(lfile);
	  lfile=0;
	}else{
	  FILE* f=file_open(line);
	  if(f){
	    read_file(f);//just read
	    fclose(f);
	  }
	}
        break;
      }
      case 'a':{
        awaiting=0;
        FILE* f=file_open(line);
        if(f){
          load_file(f);//load into memory
          fclose(f);
        }
        break;
      }
      case 'l':{
        awaiting=0;
	xltoggle=XL_LINE;
        if(lfile)free(lfile);
        lfile=strdup(line);
        break;
      }
      case 'x':{
	awaiting=0;
	xltoggle=XL_EXTENDED;
	if(lfile)free(lfile);
	lfile=strdup(line);
	break;
      }
      case 'd':{
        awaiting=0;
        delim=strdup(line);
        break;
      }
      case 't':{
        awaiting=0;
        if(lfile){//if we want to apply to another file
            FILE* l=file_open(lfile);
            if(l){
                interlace_to_file(line,l);
                fclose(l);
            }
            free(lfile);
            lfile=0;
        }
        else interlace_to_line(line);
        break;
      }
      case 'D':{
        awaiting=0;
        free(Fd);
        Fd=formatdup(line);
        break;
      }
      case 'F':{
        awaiting=0;
        free(Ff);
        Ff=formatdup(line);
        break;
      }
    }
    awaiting=0;
    if(tofree){
      free(tofree);
      tofree=0;
    }
    return;
  }

  //empty line?
  if(!*line)return;

  //ordinary line of code
  if(lfile){//applying to a file?
    FILE* l=file_open(lfile);
    if(l){
      if(xltoggle==XL_LINE){
	apply_to_file(line,l,false);
      }else{
	extended_to_file(line,l,false);
      }
      fclose(l);
    }
    free(lfile);
    lfile=0;
  }else{//just a line
    char* comment;
    array a;
    evaluate_line(line,&comment,true,a);
    if(comment){
      add_to_history(a);
    }
  }
}

int eval_control::format_splitline(char* line, vector<format_piece>& fmt){
  int countnums=0;
  char* c=line;

  format_piece p=(format_piece){'s',c,c};
  const char* delims=" \t,;:";
  while(*c && *c!='\n'){//iteration
    if(p.c=='d'){//number
      char* end;
      strtold(c,&end);//find the end of the number
      p.parsed_start=c;//this 'd' part of the condition only happens at the beginning of the number
      p.parsed_stop=end;
      if(end-c>0){
	fmt.push_back(p);//write down the number
	countnums++;
	p=(format_piece){'s',end,end};//new string
	c=end;
	continue;
      }else{
	p.c='s';
	c=end;
      }
    }else{//string
      if(isdigit(*c)||*c=='.'||*c=='-'){//found number, minus may start a number
	if(c==p.parsed_start){//a new number immediately, convert to number (this happens at the beginning)
	  p.c='d';
	  continue;
	}else{
	  if(strchr(delims,*(c-1)) ){//if previous is a delimiter
	    p.parsed_stop=c;
	    fmt.push_back(p);
	    p=(format_piece){'d',c,c};//new number
	    continue;
	  }else{//if not after delimiter, part of a string
	    //do nothing
	  }
	}
      }
    }
    c++;
  }
  //cleanup: if anything is still in the "p", add to format
  if(c>p.parsed_start){
    p.parsed_stop=c;
    fmt.push_back(p);
    if(p.c=='d')countnums++;
  }
  return countnums;
}

void eval_control::extended_to_file(char* l, FILE* f, bool file_or_line){
  vector<format_piece> fmt;

  //compile code
  list<code_chunk> code;
  int errcode=retrieve_code(l,code,file_or_line);
  if(errcode==2)return;

  int line=1;
  while(bufgets(f)){
    memory[0].clear();
    memory[0].push_back((itype)line);
    //parse. If no numbers, just return the same number and carry on
    fmt.clear();
    if(!format_splitline(buffer,fmt)){
      fputs(buffer,stdout);
      continue;
    }
    //here we fill the memory with slots from the input line
    memory[1].clear();
    for(int i=0;i<fmt.size();i++){
      format_piece fp=fmt[i];
      if(fp.c=='d'){//number
	char end=*fp.parsed_stop;
	*fp.parsed_stop=0;//fake end of string
	memory[1].push_back(value(fp.parsed_start));
	*fp.parsed_stop=end;//restore original character
      }
    }
    //evaluate the line
    array a;
    evaluate_code(code,false,a);
    line++;
    if(!errcode){
      int outsize=a.size();
      int outpos=0;
      for(int i=0;i<fmt.size();i++){
	format_piece fp=fmt[i];
	if(fp.c=='d'){//number
	  if(outpos<outsize){
	    a[outpos].print(stdout,Fd,Ff);
	    outpos++;
	  }
	}else{//string
	  char end=*fp.parsed_stop;
	  *fp.parsed_stop=0;//fake end of string
	  fputs(fp.parsed_start,stdout);
	  *fp.parsed_stop=end;//restore original character
	}
      }
      //if there's anything left
      while(outpos<outsize){
	fputs(delim,stdout);
	a[outpos].print(stdout,Fd,Ff);
	outpos++;
      }
      putc('\n',stdout);//terminate line
    }
  }

}

void eval_control::add_to_history(array& arr){
  histsize++;
  memory[-histsize]=arr;
}

eval_control::eval_control(){//initialization
#ifdef TIMERAND
  srand(time(0));
#else
  {
    uint seed;
    FILE* f=fopen("/dev/urandom","r");
    fread(&seed,sizeof(uint),1,f);
    srand(seed);
    fclose(f);
  }
#endif
  histsize=0;
  toggle_print=0;
  mempos=0;
  awaiting=0;
  lfile=0;
  xltoggle=XL_LINE;
  delim=strdup(" ");
  Fd=formatdup("%d");
  Ff=formatdup("%g");
  bufsize=BUF_SIZE;
  buffer=(char*)malloc(bufsize);
}

eval_control::~eval_control(){
  //free buffer
  free(buffer);
  //free other allocated strings
  free(delim);
  free(Fd);
  free(Ff);
}
